#!/bin/bash

# This file is a part of The Silly Little Oracle Benchmark (a.k.a. SLOB)
#
# Copyright (c) 1999-2015 Kevin Closson and Kevin Closson d.b.a. Peak Performance Systems
#
# The Software
# ------------
# SLOB is a collection of software, configuration files and documentation (the "Software").
#
# Use
# ---
# Permission is hereby granted, free of charge, to any person obtaining a copy of the Software, to
# use the Software. The term "use" is defined as copying, viewing, modifying, executing and disclosing
# information about use of the Software to third parties.
#
# Redistribution
# --------------
# Permission to redistribute the Software to third parties is not granted. The Software           
# is obtainable from kevinclosson.net/slob. Any redistribution of the Software to third parties
# requires express written permission from Kevin Closson.
#
# The copyright notices and permission notices shall remain in all copies of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
# USE OR OTHER DEALINGS IN THE SOFTWARE.
 

function msg() {
local type="$1"
local msg="$2" 
local now=$(date +"%Y.%m.%d-%H:%M:%S")

echo "${type} : ${now} : ${msg}"
}

function test_conn() {
local constring="$*"
local ret=""
local outfile=/tmp/.${RANDOM}_slob_test_conn.out

msg NOTIFY "Testing connectivity. Command: \"sqlplus -L $constring\" "

sqlplus -L "$constring" <<EOF > $outfile 2>&1
WHENEVER SQLERROR EXIT 2;
SELECT COUNT(*) FROM DUAL;
EXIT;
EOF

ret=$?

if [ $ret -ne 0 ]
then
	msg FATAL "${FUNCNAME}: Instance connectivity check failed. SQL*Plus output: "
	cat $outfile
fi

rm -f $outfile
return $ret
}

function is_host_nt() {
local constring="$1"
local tmp=""

sqlplus -s -L  "$constring" <<EOF | while read tmp
SET HEA OFF
SELECT DBMS_UTILITY.PORT_STRING FROM DUAL ;
EXIT;
EOF
do
	if ( echo $tmp | grep NT > /dev/null 2>&1 )
	then
		return 0
	fi
done

return 1
}

function check_for_stragglers() {
local string=""
local tmpfile="/tmp/.${RANDOM}_slob_monitor_processes.out"
local flag=FALSE

cat /dev/null > $tmpfile

for string in 'iostat -xm 3' 'mpstat -P ALL 3' 'vmstat 3'
do

	if ( pgrep -f "$string" >> $tmpfile )
	then
		flag=TRUE
	fi
done

if [ "$flag" = "TRUE" ]
then
	msg NOTIFY ""
	msg NOTIFY ""
	msg WARNING "**********"
	msg WARNING "SLOB has found possible zombie processes from a prior SLOB test."
	msg WARNING "It is possible that a prior SLOB test aborted. These processes may be"
	msg WARNING "appending to unlinked files which may fill up a file system."
	msg WARNING "Please investigate the following processes:"
	msg WARNING "**********"
	ps -f `cat $tmpfile`
fi

rm -f $tmpfile
}

function check_update_pct_logic() {
local pct=$1

[ $pct -lt 0 ] || [ $pct -gt 100 ] && return 1
[ $pct -gt 50 ] && [ $pct -le 99 ] && return 1

return 0
}

function check_hotspot_logic() {
local bsz=$1
local scale=$2
local hotspot_mb=$3
local offset_mb=$4
local ads_mb=0
local x=0

for x in $hotspot_mb $offset_mb
do
	if ( ! is_int "$x" )
	then
		msg FATAL "Hot Spot tunable paramters in slob.conf must be integer values"
		return 1	
	fi
done

(( ads_mb = $scale * $bsz / 1024 / 1024      ))
(( hotspot_zone_mb = $offset_mb + $hotspot_mb ))

if [ $hotspot_zone_mb -gt $ads_mb ]
then
	msg FATAL "Logic Error: Check slob.conf settings"
	msg FATAL ""
	msg FATAL "Total SLOB logical block range is ${ads_mb} MB (see slob.conf->SCALE)"
	msg FATAL "User specified hotspot offset: ${offset_mb} MB offset (see slob.conf->HOTSPOT_OFFSET_MB)"
	msg FATAL "User specified hotspot size ${hotspot_mb} MB (see slob.conf->HOTSPOT_MB)"
	msg FATAL "${offset_mb} + ${hotspot_mb} > ${ads_mb}"
	msg FATAL "Hotspot extends beyond SLOB logical block range"
	return 1
fi

return 0
}

function fix_scale() {
#global MIN_SCALE
local bsz="$1"
local scale="$2"
local factor=""

# Check for permissible values first
factor=$( echo "$scale" | sed 's/[0-9 MGT]//g'   )
[[ -n "$factor" ]] && return 1

#Work out scale:
factor=$( echo "$scale" | sed 's/[^MGT]//g'   )
if [ -z "$factor" ]
then
        # This is a simple integer assigment case
        scale=$( echo "$scale" | sed 's/[^0-9]//g' )
else
        scale=$( echo "$scale" | sed 's/[^0-9]//g' )
        case "$factor" in
                "M") (( scale = ( 1024 * 1024 * $scale  ) / $bsz ))  ;;
                "G") (( scale = ( 1024 * 1024 * 1024 * $scale  ) / $bsz )) ;;
                "T") (( scale = ( 1024 * 1024 * 1024 * 1024 *  $scale  ) / $bsz )) ;;
                * ) return 1 ;;
        esac

fi

if ( ! is_int $scale )
then
	msg FATAL "${FUNCNAME}: Computed is \"${scale}\". Please report this logic error."
	return 1
fi

if [ $scale -lt $MIN_SCALE ]
then
	echo $scale
	return 1
else

	echo $scale
	return 0
fi

}
function validate_scale() {
local requested_scale=$1
local num_schemas=$2
local user="$3"
local ret=0
local outfile=/tmp/.${RANDOM}_slob_validate_scale.out

cat /dev/null > $outfile 

sqlplus $user/$user <<EOF 2>/dev/null | sed 's/^.* FATAL/FATAL/g' | grep FATAL > $outfile
SET SERVEROUTPUT ON   ;
SET VERIFY OFF;

VARIABLE exit_status NUMBER;

DECLARE

v_requested_test_scale PLS_INTEGER := $requested_scale ;
v_num_schemas PLS_INTEGER := $num_schemas ;

v_sql VARCHAR2(80);
v_rows NUMBER;
v_curr_schema PLS_INTEGER := 1;

BEGIN

:exit_status := 0 ;

FOR v_curr_schema in 1..v_num_schemas LOOP
	v_sql := 'ALTER SESSION SET CURRENT_SCHEMA = user' || v_curr_schema ;
	EXECUTE IMMEDIATE v_sql;

	SELECT MAX(custid) INTO v_rows FROM CF1;

	IF ( v_requested_test_scale >  v_rows ) THEN

		dbms_output.put_line('FATAL:                       : SLOB schema user' || v_curr_schema || ' has ' || v_rows || ' rows.' );
		dbms_output.put_line('FATAL:                       : User specified scale: ' || v_requested_test_scale );
		dbms_output.put_line('FATAL:                       : Cannot test more data than has been loaded.' );

		:exit_status := v_curr_schema;
		EXIT ;
	END IF;
END LOOP;
END;
/

EXIT :exit_status ;
EOF


if ( grep FATAL $outfile > /dev/null 2>&1 )
then
	msg FATAL ""
	msg FATAL "Invalid slob.conf setting for SCALE parameter"
	msg FATAL ""
	cat $outfile
	rm -f $outfile
	return 1
else
	rm -f $outfile
	return 0
fi


}

function count_pids() {
local pidfile=$1
local numpids=0

ps -p `cat $pidfile` | wc -l
return 0

}

function wait_pids() {
local sessions=$1
local run_time=$2
local wl="$3"
local pids="$4"
local sleeptm=$(( run_time - 3 ))
local cnt=0
local monitor_limit=0
local x=0
local tmp=""
local pid_file="/tmp/.${RANDOM}_slob_pids.out"

echo "$pids" > $pid_file 2>&1

if [ "$wl" -eq 0 ]
then
	msg NOTIFY "List of monitored sqlplus PIDs written to $pid_file"

	monitor_limit=300

	sleep 5
	tmp=`count_pids $pid_file`

	if [ $tmp -lt $sessions ]
	then
		msg FATAL "SLOB process monitoring discovered $(( sessions - tmp )) sqlplus processes have aborted."
		msg FATAL "Consider export SLOB_DEBUG=TRUE, re-running the test and then examine slob_debug.out"
		rm -f $pid_file
		return 1
	fi

	msg NOTIFY "Waiting for $(( sleeptm - 5 )) seconds before monitoring running processes (for exit)."
	sleep $(( sleeptm - 5 ))
else
	msg NOTIFY "This is a fixed-iteration run (see slob.conf->WORK_LOOP). "
	monitor_limit=0
fi

msg NOTIFY "Entering process monitoring loop."

while ( ps -p $pids > /dev/null 2>&1 )
do
	
	if [ "$monitor_limit" -ne 0 ]
	then
		if [ "$cnt" -gt $monitor_limit ]
		then
			msg FATAL "The following SQL*Plus processes have not exited after $monitor_limit seconds."
			ps -fp $pids
			return 1
		fi
	fi
	(( cnt = $cnt + 1 ))
	(( x = $cnt % 10 ))
	if [ $x = 0 ]
	then
		tmp=`count_pids $pid_file`		
		msg NOTIFY "There are $tmp sqlplus processes remaining."
	fi	

	sleep 1
done
rm -f $pid_file

return 0
}

function snap_awr() {
local admin_conn="$1"
local snap=0

snap=`$admin_conn <<EOF
SET HEADING OFF
SET FEEDBACK OFF
SET PAGES 0

EXEC DBMS_WORKLOAD_REPOSITORY.CREATE_SNAPSHOT
SELECT max(SNAP_ID) FROM dba_hist_snapshot;
EXIT;
EOF
`
echo $snap | sed 's/[^0-9]//g'
}

function compress_file() {
local f="$1"

if [ ! -f $f ]
then
	msg NOTIFY "Cannot compress file \"$f\" "
	msg NOTIFY "No such file."
	return 1
else
	gzip -9 $f
fi

return 0
}


function run_awr() {
local admin_conn="$1"
local begin_snap="$2"
local end_snap="$3"
local tmp=""

($admin_conn <<EOF
set echo off heading on underline on;
column inst_num  heading "Inst Num"  new_value inst_num  format 99999;
column inst_name heading "Instance"  new_value inst_name format a12;
column db_name   heading "DB Name"   new_value db_name   format a12;
column dbid      heading "DB Id"     new_value dbid      format 9999999999 just c;

column end_snap new_value end_snap;
column begin_snap new_value begin_snap;


select d.dbid            dbid
     , d.name            db_name
     , i.instance_number inst_num
     , i.instance_name   inst_name
  from v\$database d,
       v\$instance i;

define  begin_snap = '$begin_snap' ;
define  end_snap = '$end_snap' ;
define  num_days     = 1;
define  report_type  = 'html';
define  report_name  = 'awr.html';

@?/rdbms/admin/awrrpti
exit
EOF
) &

($admin_conn <<EOF

set echo off heading on underline on;
column inst_num  heading "Inst Num"  new_value inst_num  format 99999;
column inst_name heading "Instance"  new_value inst_name format a12;
column db_name   heading "DB Name"   new_value db_name   format a12;
column dbid      heading "DB Id"     new_value dbid      format 9999999999 just c;

column end_snap new_value end_snap;
column begin_snap new_value begin_snap;


select d.dbid            dbid
     , d.name            db_name
     , i.instance_number inst_num
     , i.instance_name   inst_name
  from v\$database d,
       v\$instance i;

define  begin_snap = '$begin_snap' ;
define  end_snap = '$end_snap' ;

define  num_days     = 1;

define  report_type  = 'text';
define  report_name  = 'awr.txt';

@?/rdbms/admin/awrrpti

define  begin_snap = '$begin_snap' ;
define  end_snap = '$end_snap' ;
define  report_type  = 'text';
define  report_name = 'awr_rac.txt';

@?/rdbms/admin/awrgrpt.sql

define  begin_snap = '$begin_snap' ;
define  end_snap = '$end_snap' ;
define  report_type  = 'html';
define  report_name  = 'awr_rac.html';

@?/rdbms/admin/awrgrpt.sql

exit;

EOF
) &

wait

for tmp in awr.html awr_rac.html
do
	msg NOTIFY "Compressing file \"$tmp\" "
	( compress_file $tmp ) &
done
wait
return 0

}


function print_usage() {

echo "
${0} supports the following command usage:

1. Single Option Invocation.
        $ sh ${0} <number-of-SLOB-schemas-to-test>

2. Multiple Option Invocation 
   2.1 This invocation style requires *exactly* four options.
        $ sh ${0} -s <number-of-slob-schemas-to-test> -t <SLOB-threads-per-schema>

NOTE: With Single Option Invocation slob.conf->THREADS_PER_SCHEMA is used. If you
      want more than a single SLOB thread per schema set THREADS_PER_SCHEMA in slob.conf.
      The default setting for slob.conf->THREADS_PER_SCHEMA is 1.
      
      With Multiple Option Invocation slob.conf->THREADS_PER_SCHEMA is overridden.
      The number of SLOB threads per schema is taken from the argument passed 
      in with the -t option.

EXAMPLES:

      Example 1. 256 SLOB schemas each with slob.conf->THREADS_PER_SCHEMA number 
      of SLOB threads per schema:
        $ sh ${0} 256

      Example 2. 16 SLOB schemas each with 32 SLOB threads:
        $ sh ${0} -s 16 -t 32

NOTE: Example 2 produces 512 (16*32) Oracle Database sessions.

ADDITIONAL INFORMATION: The SLOB documentation at kevinclosson.net/slob or SLOB/doc 

"
return 0
}

function is_int() {
local s="$1"

if ( ! echo $s |  grep -q "^-\?[0-9]*$" ) 
then
	return 1
else
	return 0
fi
}

function check_cmdline() {
#global SCHEMAS
local args="$*"
local tmp=""
local i=0

# The option/args must be either:
#     A single option that is an integer representing the number of schemas to test
# or
#     4 tokens: A -s option with integer arg **and** -t option with an integer arg
#
# The this routine either echoes a value for SCHEMAS and returns or prepares the cmdline for
# getopts processing of multiple option invocation.

[[ $# -ne 1 && $# -ne 4 ]]  && return 1 # Force 1 or 4

if [[ $# -eq 1 ]] # Quickly handle single option invocation
then

	if ( ! is_int "$1" ) 
	then
		msg FATAL "Single Option Invocation requires one option: A non-zero integer"
		return 1
	fi

	if [[ "$1" -lt 1 ]]
	then
		msg FATAL "Single Option Invocation requires one option: A non-zero integer"
		return 1
	fi
	
	echo "$1"
	return 0
fi

# Deal with multiple option invocation.
# There are the requisite number of options greater than 1 and 
# now the first and third arg must be suitable for getopts processing

for tmp in $1 $3
do
	if ( ! echo "$tmp" | grep -q '\-' > /dev/null 2>&1 )
	then
		msg FATAL "Invalid Multiple Option Invocation option/argument pairs."
		return 1	
	fi		
done

return 0
}



#---------- Main body

SCHEMAS=""
CMDLINE_THREADS_PER_SCHEMA=""
opt=""
tmp=""
ret=0


tmp=$(check_cmdline $*) ; ret=$?   # either set SCHEMAS or validate the multiple option invocation cmdline

if [ $ret -ne 0 ]
then
	echo "$tmp"
	msg FATAL "Invalid command line. Abort."
	print_usage
	exit 1
else

	SCHEMAS="$tmp"
fi

while getopts ":s:t:" opt 
do
	case "$opt" in
		s)
			SCHEMAS="$OPTARG"
			if ( ! is_int $SCHEMAS || [[ "$SCHEMAS" -lt 1 ]] )
			then
				msg FATAL "The -s option requires a non-zero integer argument"
				print_usage 
				exit 1
			fi
		;;
		t)
			tmp="$OPTARG"
			if ( ! is_int $tmp || [[ "$tmp" -lt 1 ]] )
			then
				msg FATAL "The -t option requires a non-zero integer argument"
				print_usage 
				exit 1
			else

				export CMDLINE_THREADS_PER_SCHEMA=$tmp
			fi
		;;
		*) 
			print_usage
			exit 1
		;;	
	esac
done

source ./slob.conf

# BLOCK_SZ Can be set in the environment or slob.conf. Not documented.
# SLOB supports all Oracle block sizes

BLOCK_SZ=${BLOCK_SZ:=8192}
MIN_SCALE=10000

# Just in case user deleted lines in slob.conf:

UPDATE_PCT=${UPDATE_PCT:=25}
RUN_TIME=${RUN_TIME:=300}
WORK_LOOP=${WORK_LOOP:=0}
SCALE=${SCALE:=10000}
WORK_UNIT=${WORK_UNIT:=256}
REDO_STRESS=${REDO_STRESS:=LITE}
HOT_SCHEMA_FREQUENCY=${HOT_SCHEMA_FREQUENCY:=0}
THREADS_PER_SCHEMA=${THREADS_PER_SCHEMA:=1}

DO_HOTSPOT=${DO_HOTSPOT:=FALSE}
HOTSPOT_MB=${HOTSPOT_MB:=10}
HOTSPOT_OFFSET_MB=${HOTSPOT_OFFSET_MB:=0}
HOTSPOT_FREQUENCY=${HOTSPOT_FREQUENCY:=0}

THINK_TM_FREQUENCY=${THINK_TM_FREQUENCY:=0}
THINK_TM_MIN=${THINK_TM_MIN:=.1}
THINK_TM_MAX=${THINK_TM_MAX:=.5}

ADMIN_SQLNET_SERVICE=${ADMIN_SQLNET_SERVICE:=''}
SQLNET_SERVICE_BASE=${SQLNET_SERVICE_BASE:=''}
SQLNET_SERVICE_MAX=${SQLNET_SERVICE_MAX:=''}
SYSDBA_PASSWD=${SYSDBA_PASSWD:=''}

# Handle command line overrides of slob.conf settings
if  [[ -n "$CMDLINE_THREADS_PER_SCHEMA" ]] 
then
	export THREADS_PER_SCHEMA=$CMDLINE_THREADS_PER_SCHEMA
fi

sqlplus_pids=""
misc_pids=""
before=""
tm=""
cmd=""
nt=0
do_rotor="FALSE"
slobargs=""
sleep_secs=""
cnt=1
x=0
instance=1
connect_string=""
ret=0
tmp=""
user_scale_value=""

export non_admin_connect_string=""
export admin_connect_string="/ as sysdba"

if [ "$SLOB_DEBUG" = "TRUE" ]
then
	export debug_outfile="slob_debug.out"
	msg NOTIFY "Debug info being sent to $debug_outfile "
else
	export debug_outfile="/dev/null"
fi

if [ ! -x ./mywait ]
then
	msg FATAL " "
	msg FATAL "./mywait executable not found or wrong permissions."
	msg FATAL "Please change directories to ./wait_kit and execute make(1)."
	msg NOTIFY "Example: "
	msg NOTIFY " $ cd ./wait_kit "	
	msg NOTIFY " $ make "	
	msg FATAL "SLOB abnormal end."
	exit 1
fi

if ( ! type sqlplus  >> $debug_outfile 2>&1 )
then
	msg FATAL "sqlplus is not executable in $PATH"
	msg FATAL "SLOB abnormal end."
	exit 1
fi

# Enforce maximum tested database sessions
if [[ $(( SCHEMAS * THREADS_PER_SCHEMA )) -gt 4096 ]]
then
	msg FATAL "Cannot attach ${THREADS_PER_SCHEMA} (slob.conf) sessions to $SCHEMAS schemas"
	msg FATAL "Maximum tested active database session count is 4096."
	msg FATAL "Abort"
	exit 1
fi

rm -f  awr.txt awr.*html awr*.gz iostat.out vmstat.out mpstat.out db_stats.out sqlplus.out slob_debug.out  
rm -f /tmp/.[0-9]*_slob*.out 


# The following is the first screen output
msg NOTIFY " "
msg NOTIFY "Conducting SLOB pre-test checks."


if [ -n "$ADMIN_SQLNET_SERVICE" ]
then
        export admin_connect_string="sys/${SYSDBA_PASSWD}@${ADMIN_SQLNET_SERVICE} as sysdba"
        export non_admin_connect_string="@${SQLNET_SERVICE_BASE}"
fi

if [ -n "$SQLNET_SERVICE_BASE" ]
then
	if [[ -n "$SQLNET_SERVICE_MAX" && $SQLNET_SERVICE_MAX -gt 0 ]]
	then
		do_rotor=TRUE	
		(( sqlnet_batch = $SCHEMAS / $SQLNET_SERVICE_MAX ))

		if [ "$SQLNET_SERVICE_MAX" -eq 1 ]
		then
			# User wants all SLOB sessions connected through ${SQLNET_SERVICE_BASE}${SQLNET_SERVICE_MAX}  (e.g., RAC1)
			msg NOTIFY "All SLOB sessions will connect to ${SQLNET_SERVICE_BASE}${SQLNET_SERVICE_MAX} via SQL*Net "
		else
			# User wants round-robin from 1 through N (e.g., RAC1,RAC2,RAC3,RAC4)
			msg NOTIFY "SLOB sessions will connect round-robin from ${SQLNET_SERVICE_BASE}1 through ${SQLNET_SERVICE_BASE}${SQLNET_SERVICE_MAX} via SQL*Net"
			msg NOTIFY "Instances 1 through ${SQLNET_SERVICE_MAX} will each have $sqlnet_batch sessions connected."
		fi
	else
		# User wants all SLOB sessions connected through SQLNET_SERVICE_BASE (e.g., RAC)
		msg NOTIFY "All SLOB sessions will connect to ${SQLNET_SERVICE_BASE} via SQL*Net"
	fi
fi

admin_conn="sqlplus -L $admin_connect_string"
admin_conn_silent="sqlplus -S -L $admin_connect_string"


# Use fix_scale to convert any legal value to number of blocks
user_scale_value="$SCALE"
SCALE=$(fix_scale "$BLOCK_SZ" "$user_scale_value") ; ret=$?

if [ $ret -ne 0 ]
then
        msg FATAL "The value assigned to slob.conf->SCALE ($user_scale_value [$SCALE]) is an illegal value. "
        msg FATAL "Illegal SCALE value. Mininum value is $MIN_SCALE blocks"
        msg FATAL "Abort"
        exit 1
fi

# Check to make sure the schemas have enough data to cover the testing
if ( ! validate_scale $SCALE $SCHEMAS user1 )
then
	msg FATAL ""
	msg FATAL "User specified testing SCALE $SCALE per schema."
	msg FATAL "Insufficient data has been loaded to test at scale $SCALE (per schema)."
	msg FATAL ""
	exit 1
fi

# Enhance output to show command line overrides

disp1="THREADS_PER_SCHEMA: $THREADS_PER_SCHEMA"

if [ -n "$CMDLINE_THREADS_PER_SCHEMA" ]
then
	disp1="THREADS_PER_SCHEMA: $THREADS_PER_SCHEMA (-t option)"
fi



echo " 
UPDATE_PCT: $UPDATE_PCT			
RUN_TIME: $RUN_TIME
WORK_LOOP: $WORK_LOOP			
SCALE: $user_scale_value ($SCALE blocks)
WORK_UNIT: $WORK_UNIT			
REDO_STRESS: $REDO_STRESS
HOT_SCHEMA_FREQUENCY: $HOT_SCHEMA_FREQUENCY
DO_HOTSPOT: $DO_HOTSPOT			
HOTSPOT_MB: $HOTSPOT_MB
HOTSPOT_OFFSET_MB: $HOTSPOT_OFFSET_MB
HOTSPOT_FREQUENCY: $HOTSPOT_FREQUENCY	
THINK_TM_FREQUENCY: $THINK_TM_FREQUENCY
THINK_TM_MIN: $THINK_TM_MIN		
THINK_TM_MAX: $THINK_TM_MAX

${disp1}

ADMIN_SQLNET_SERVICE: \"$ADMIN_SQLNET_SERVICE\"
SQLNET_SERVICE_BASE: \"$SQLNET_SERVICE_BASE\"
SQLNET_SERVICE_MAX: \"$SQLNET_SERVICE_MAX\"

Connect strings to be used:
admin_connect_string: \"$admin_connect_string\"
non_admin_connect_string: \"$non_admin_connect_string\"
admin_conn: \"$admin_conn\"
"

for tmp in $UPDATE_PCT $RUN_TIME $WORK_LOOP $WORK_UNIT $HOT_SCHEMA_FREQUENCY $HOTSPOT_MB $HOTSPOT_OFFSET_MB $HOTSPOT_FREQUENCY 
do
	if ( ! is_int $tmp )
	then
		msg FATAL "These slob.conf parameters must be assigned integer values:
		UPDATE_PCT RUN_TIME WORK_LOOP WORK_UNIT HOT_SCHEMA_FREQUENCY HOTSPOT_MB HOTSPOT_OFFSET_MB HOTSPOT_FREQUENCY "
		msg FATAL "\"$tmp\" is not an integer. Please check slob.conf."
		exit 1
	fi
done


if ( ! check_update_pct_logic $UPDATE_PCT )
then
	msg WARNING "*************************************************************************"
	msg WARNING "Values between 51 and 99 for UPDATE_PCT render non-deterministic results."
	msg WARNING "The slob.conf value assigned to UPDATE_PCT is $UPDATE_PCT"
	msg WARNING ""
	msg WARNING "Please consider values in the following ranges: 0-50 or 100 for optimal"
	msg WARNING "test repeatability."
	msg WARNING "*************************************************************************"
fi


# Check Hotspot settings for sanity
if [ "$DO_HOTSPOT" = "TRUE" ]
then
	if ( ! check_hotspot_logic $BLOCK_SZ $SCALE $HOTSPOT_MB $HOTSPOT_OFFSET_MB )
	then
		msg FATAL "Abort"
		exit 1
	fi
fi

msg NOTIFY "Testing SYSDBA connectivity to the instance to validate slob.conf settings."

if ( ! test_conn ${admin_connect_string} )
then
        msg FATAL "Cannot connect to the instance."
	msg FATAL "Connect string: \"${admin_connect_string}\""
        msg FATAL "Please verify the instance and listener are running and the settings"
        msg FATAL "in slob.conf are correct for your connectivity model."
	msg FATAL ""
	msg FATAL "If not connecting with SQL*Net please check \$ORACLE_SID."
	msg FATAL "Current value for \$ORACLE_SID is \"$ORACLE_SID\""
	msg FATAL ""
	
	msg FATAL "SLOB abnormal end."


        exit 1
fi

# The following section tests test if the schemas exist and can be accessed via SQL*Net as per
# the settings in slob.conf
#

if [ "$do_rotor" = "TRUE" ]
then

	if [[ $(( SCHEMAS % SQLNET_SERVICE_MAX )) != 0 ]]
	then
		msg WARNING "**********"
		msg WARNING "Schema count ($SCHEMAS) is not an even multiple of ${SQLNET_SERVICE_MAX}. Sessions per instance will not by symmetrical."
		msg WARNING "**********"
	fi

	msg NOTIFY "Connections will rotor from ${non_admin_connect_string}1 through ${non_admin_connect_string}${SQLNET_SERVICE_MAX}"

	for (( i = 1 ; i <= $SQLNET_SERVICE_MAX ; i++ ))
	do
		connect_string="${non_admin_connect_string}${i}"

		for U in 1 ${SCHEMAS} 
		do
			#msg NOTIFY "Testing connectivity to instance (user${U}/user${U}${connect_string})"

			if ( !  test_conn "user${U}/user${U}${connect_string}" )
			then
				msg FATAL "Connect failure user${U}/user${U}${connect_string}"
				msg FATAL "SLOB abnormal end."
				exit 1
			fi
		done
	done

else

	for U in 1 ${SCHEMAS} 
	do
		#msg NOTIFY "Testing connectivity to instance (user${U}/user${U}${non_admin_connect_string})"	
		if ( !  test_conn "user${U}/user${U}${non_admin_connect_string}" )
		then

			msg FATAL "Connect failure user${U}/user${U}${non_admin_connect_string}"
			msg FATAL "SLOB abnormal end."
			exit 1
		fi

		[[ $SCHEMAS -eq 1 ]] && break
	done
fi


if (  is_host_nt "$admin_connect_string" )
then
	msg NOTIFY "Database host operating system is NT. The db_stats.out file will not be produced."
	nt=1
else
	nt=0
fi

msg NOTIFY "Performing redo log switch."
sqlplus $admin_connect_string @./misc/switchlog > /dev/null 2>&1
msg NOTIFY "Redo log switch complete."

msg NOTIFY "Setting up trigger mechanism."
./create_sem > /dev/null 2>&1

if [  -z "$NO_OS_PERF_DATA" ]
then
	msg NOTIFY "Running iostat, vmstat and mpstat on current host--in background."
	( iostat -xm 3 > iostat.out 2>&1 ) &
	misc_pids="${misc_pids} $!"
	( vmstat 3 > vmstat.out 2>&1 ) &
	misc_pids="${misc_pids} $!"
	( mpstat -P ALL 3  > mpstat.out 2>&1) &
	misc_pids="${misc_pids} $!"
fi

msg NOTIFY "Connecting ${THREADS_PER_SCHEMA} sessions to ${SCHEMAS} schema(s) ..."

#
# Launch the sessions
#


cnt=1 ; x=0 ; spawn_throttle=0 ; instance=1  

until [ $cnt -gt $SCHEMAS ]
do
	if [ "$do_rotor" = "TRUE" ]
	then
		cmd="sqlplus -s user${cnt}/user${cnt}${non_admin_connect_string}${instance}"
		(( instance = $instance + 1 ))
		[[ $instance -gt $SQLNET_SERVICE_MAX ]] && instance=1
	else
		cmd="sqlplus -s user${cnt}/user${cnt}${non_admin_connect_string}"
	fi

	arg1=$cnt
	arg2=$UPDATE_PCT
	arg3=$WORK_LOOP
	arg4=$RUN_TIME
	arg5=$SCALE 
	arg6=$WORK_UNIT 
	arg7=$REDO_STRESS 
	arg8=$HOT_SCHEMA_FREQUENCY 
	arg9=$nt 
	arg10=$DO_HOTSPOT 
	arg11=$HOTSPOT_MB 
	arg12=$HOTSPOT_OFFSET_MB
	arg13=$HOTSPOT_FREQUENCY 
	arg14=$THINK_TM_FREQUENCY 
	arg15=$THINK_TM_MIN 
	arg16=$THINK_TM_MAX

	slobargs="$arg1 $arg2 $arg3 $arg4 $arg5 $arg6 $arg7 $arg8 $arg9 $arg10 $arg11 $arg12 $arg13 $arg14 $arg15 $arg16"

	for ((i=0 ; i < $THREADS_PER_SCHEMA ; i++))
	do
		(( spawn_throttle = $spawn_throttle + 1 ))
		( $cmd @slob $slobargs >> $debug_outfile 2>&1 ) &
		sqlplus_pids="${sqlplus_pids} $!"

		(( x = $spawn_throttle % 17 ))
		[[ $x -eq 0 ]] && spawn_throttle=0 && sleep .5
	done
	(( cnt = $cnt + 1 ))
done

msg NOTIFY " "

if [[  $(( cnt / 6 )) -gt 5 ]]
then
	sleep_secs=5
else
	(( sleep_secs = $cnt / 6 ))
fi

if [ "$sleep_secs" -gt 1 ]
then
	msg NOTIFY "Pausing for $sleep_secs seconds before triggering the test."
	sleep $sleep_secs
fi

# Switch output to fd2 so $misc_pids shell kill feedback falls in line

msg NOTIFY "Executing AWR \"before snap\" procedure. Connect string is \"$admin_conn_silent\"  " >&2
begin_snap=`snap_awr "$admin_conn_silent"`

msg NOTIFY " "
msg NOTIFY "Triggering the test." >&2
before=$SECONDS


./trigger > /dev/null 2>&1

if ( ! wait_pids "$SCHEMAS" "$RUN_TIME" "$WORK_LOOP" "$sqlplus_pids" )
then
	msg FATAL "This is not a successful SLOB test." >&2
	exit 1
fi

(( tm =  $SECONDS - $before ))

msg NOTIFY "Run time in seconds was:  $tm" >&2
echo "Tm $tm" > tm.out

msg NOTIFY "Executing AWR \"after snap\" procedure. Connect string is \"$admin_conn_silent\" " >&2

end_snap=`snap_awr "$admin_conn_silent"`

( run_awr "$admin_conn" $begin_snap $end_snap >> $debug_outfile ) &
msg NOTIFY "Generating AWR reports. HTML reports will be compressed. Connect string is \"$admin_conn\" " >&2
sleep 1

msg NOTIFY "Terminating background data collectors." >&2
/bin/kill -9 $misc_pids > /dev/null 2>&1 

wait

mv /tmp/db_stats.out . > /dev/null 2>&1
msg NOTIFY ""
msg NOTIFY "SLOB test is complete."

check_for_stragglers

exit 0




