#!/bin/bash
# This file is a part of The Silly Little Oracle Benchmark (a.k.a. SLOB)
#
# Copyright (c) 1999-2015 Kevin Closson and Kevin Closson d.b.a. Peak Performance Systems
#
# The Software
# ------------
# SLOB is a collection of software, configuration files and documentation (the "Software").
#
# Use
# ---
# Permission is hereby granted, free of charge, to any person obtaining a copy of the Software, to
# use the Software. The term "use" is defined as copying, viewing, modifying, executing and disclosing
# information about use of the Software to third parties.
#
# Redistribution
# --------------
# Permission to redistribute the Software to third parties is not granted. The Software           
# is obtainable from kevinclosson.net/slob. Any redistribution of the Software to third parties
# requires express written permission from Kevin Closson.
#
# The copyright notices and permission notices shall remain in all copies of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
# USE OR OTHER DEALINGS IN THE SOFTWARE.

function create_file() {
local f="$1"

if ( ! cat /dev/null > $f )
then
	msg FATAL ""
	msg FATAL "Cannot create file $f "
	msg FATAL ""
	return 1	
fi

if [ ! -w $f ]
then
	msg FATAL ""
	msg FATAL "File $f is not a writable file"
	msg FATAL ""
	return 1
fi

return 0
}

function check_wait_kit() {
local f=""

for f in mywait trigger create_sem
do
	if [ ! -x ./$f ]
	then
		msg NOTIFY "Please do not forget to compile the wait kit."
		msg NOTIFY "Please change directories to ./wait_kit and execute make(1)."
		msg NOTIFY "Example: "
		msg NOTIFY " $ cd ./wait_kit "
		msg NOTIFY " $ make "
		return 1
	fi
done
return 0
}

function msg() {
local msgtype="$1"
local msgtxt="$2"
local now=$(date +"%Y.%m.%d-%H:%M:%S")

	type=`echo $msgtype | awk '{ printf("%-7s\n",$1) }'`
	echo "${type} : ${now} : ${msgtxt}"

return 0
}

function is_int() {
local s="$1"

if ( ! echo $s |  grep -q "^-\?[0-9]*$" )
then
	return 1
else
	return 0
fi
}


function fix_scale() {
#global MIN_SCALE
local bsz="$1"
local scale="$2"
local factor=""

# Check for permissible values first
factor=$( echo "$scale" | sed 's/[0-9 MGT]//g'   )
[[ -n "$factor" ]] && return 1

#Work out scale:
factor=$( echo "$scale" | sed 's/[^MGT]//g'   )
if [ -z "$factor" ]
then
        # This is a simple integer assigment case
        scale=$( echo "$scale" | sed 's/[^0-9]//g' )
else
        scale=$( echo "$scale" | sed 's/[^0-9]//g' )
        case "$factor" in
                "M") (( scale = ( 1024 * 1024 * $scale  ) / $bsz ))  ;;
                "G") (( scale = ( 1024 * 1024 * 1024 * $scale  ) / $bsz )) ;;
                "T") (( scale = ( 1024 * 1024 * 1024 * 1024 *  $scale  ) / $bsz )) ;;
                * ) return 1 ;;
        esac

fi

if ( ! is_int $scale )
then
	msg FATAL "${FUNCNAME}: Computed is \"${scale}\". Please report this logic error."
	return 1
fi

if [ $scale -lt $MIN_SCALE ]
then
	echo $scale
	return 1
else

	echo $scale
	return 0
fi

}

function flag_abort() {
#global LOG
local f=$1

msg FATAL ""
msg FATAL "${FUNCNAME}: Triggering abort"
msg FATAL ""

if ( ! cat /dev/null >  $f >> $LOG 2>&1 )
then
	msg FATAL ""
	msg FATAL "${FUNCNAME}: failed to trigger abort"
	msg FATAL ""
	return 1
fi

return 0
}

function check_abort_flag() {
local f=$1

if [ -f $f ]
then
	msg FATAL ""
	msg FATAL "${FUNCNAME}: discovered abort flag"
	msg FATAL ""
	return 0
fi

return 1
}

function test_oracle_utilities () {
#global LOG
local exe=""

for exe in sqlplus tnsping
do 
	if ( ! type $exe >> $LOG 2>&1 )
	then
		msg FATAL ""
		msg FATAL "${FUNCNAME}: Please validate your environment. SQL*Plus is not executable in current \$PATH"
		msg FATAL ""
		return 1
	fi
done

return 0
}

function test_conn() {
local constring="$*"
local ret=0

msg NOTIFY "Test connectivity with: \"sqlplus -L $constring\""

sqlplus -L "$constring" <<EOF
SET TERMOUT ON
WHENEVER SQLERROR EXIT 2;
SET ECHO ON
PROMPT Performing connectivity and database state check
SELECT COUNT(*) FROM dual;
EXIT;
EOF

ret=$?
if [ $ret -ne 0 ]
then
	msg FATAL
	msg FATAL "${FUNCNAME}: Failed connectivity check"
	msg FATAL
fi

return $ret
}

function check_mto() {
local constring="$1"
local cdb=""

cdb=`sqlplus -S / as sysdba 2>&1 <<EOF
SET HEAD OFF
SELECT CDB FROM V\\$DATABASE ;
EOF
`

if ( echo "$cdb" | grep -q "YES" > /dev/null 2>&1 )
then
	return 1
fi

return 0

}

function test_listener() {
#global LOG
local svc="$1"

if ( ! tnsping $svc >> $LOG 2>&1 )
then
	msg FATAL ""
	msg FATAL "${FUNCNAME}: tnsping failed to validate SQL*Net service ( $svc )"
	msg FATAL ""
	return 1
fi

return 0
}


function test_tablespace() {
local constring="$1"
local tablespace="$2"
local ret=0

msg NOTIFY "Testing user-specified tablespace suitability"

sqlplus -L "$constring" <<EOF 2>&1
SET TERMOUT ON
SET ECHO ON
PROMPT
PROMPT The following is a test of the user-specified tablespace
PROMPT Performing the following tasks: 
PROMPT 1) Drop SIMPLE_SLOB_TEST if exists (if it doesn't exist ORA-00942 is expected)
PROMPT 2) Create table SIMPLE_SLOB_TEST in tablespace $tablespace (user-specified)
PROMPT 3) Drop freshly created SIMPLE_SLOB_TEST table
PROMPT
DROP TABLE SIMPLE_SLOB_TEST PURGE;
WHENEVER SQLERROR EXIT 2;
CREATE TABLE SIMPLE_SLOB_TEST(c number)  TABLESPACE $tablespace ;
DROP TABLE SIMPLE_SLOB_TEST PURGE;
EXIT;
EOF

ret=$?

if [ $ret -ne 0 ]
then
	msg FATAL
	msg FATAL "${FUNCNAME}: Failed tablespace suitability test"
	msg FATAL
else
	msg NOTIFY "Finished testing user-specified tablespace suitability"

fi

return $ret

}

function slob_tabs_report() {
#global WORK_DIR
#global SCALE
local constring="$*"
local ret=0
local outfile="${WORK_DIR}/slob_data_load_summary.txt"

if ( ! create_file $outfile )
then
	msg FATAL ""
	msg FATAL "${FUNCNAME}: Cannot create ${outfile}"
	msg FATAL ""
	return 1
fi

sqlplus -L "$constring" <<EOF 2>&1
WHENEVER SQLERROR EXIT 2;
SET TERMOUT ON
SET ECHO ON
SPOOL $outfile 
REM Values from ALL_TABLES are estimates unless full table
REM statistics are gathered. Values in the NUM_ROWS column should be
REM quite close to $SCALE to indicate the data loading succeeded.
REM
REM The SLOB tables should have $SCALE rows loaded as per slob.conf->SCALE
REM To get a precise number of rows for each schema one can perform:
REM SELECT MAX(custid) FROM cf1;
REM
SET LINESIZ 80
SET PAGESIZE 50000
COLUMN OWNER FORMAT a10
COLUMN TABLE_NAME FORMAT a10
COLUMN NUM_ROWS FORMAT 999,999,999,999
COLUMN BLOCKS FORMAT 999,999,999,999


WHENEVER SQLERROR EXIT 2;
SELECT OWNER, TABLE_NAME, NUM_ROWS, BLOCKS 
FROM ALL_TABLES WHERE TABLE_NAME LIKE 'CF1'
ORDER BY OWNER;
EOF

ret=$?
if [ $ret -ne 0 ]
then
	msg FATAL
	msg FATAL "${FUNCNAME}: Failed to report on ALL_TABLES"
	msg FATAL
fi

return $ret
}

function drop_users(){
#global WORK_DIR
#global MAX_SLOB_SCHEMAS
#gloabl LOG
local constring="$1"
local fname="${WORK_DIR}/drop_users.log"
local num_processed=0
local x=0

if ( ! create_file $fname )
then
	msg FATAL ""
	msg FATAL "${FUNCNAME}: Cannot create file ($fname)"
	msg FATAL ""
	return 1
fi

for (( x=0 ; x < ( MAX_SLOB_SCHEMAS + 1 )  ; x++ ))
do
        echo "SET ECHO ON 
	      DROP USER user${x} CASCADE ;"
done | sqlplus -s "$constring" 2>&1 | tee -a $fname |  grep -i "dropped" | wc -l  | while read num_processed
do
	if ( ! is_int "$num_processed" )
	then
		echo "REM WARNING: ${FUNCNAME}: Logic error (\"${num_processed}\" is not an integer value). Please report this warning." >> $fname
	else
		if [ $num_processed -gt 0 ]
		then
			msg NOTIFY "Deleted $(( num_processed - 1 )) SLOB schema(s)."
		fi
	fi
done

return 0
}

function gather_stats() {
local user=$1
local pass=$2
local degree=$3
local constring="$4"
local ret=0
sql=""

sql="ALTER TABLE cf1 PARALLEL $degree "
if ( ! run_ddl $user $pass "$constring" "$sql" )
then
       msg FATAL "${FUNCNAME}: SQL execution failed in $user schema: \"${sql}\" "
       return 1
fi

msg NOTIFY "Gathering stats on user ${user} schema"

sqlplus ${user}/${pass}${constring} <<EOF
SET TERMOUT ON
WHENEVER SQLERROR EXIT 2;
SET ECHO ON
SET TIMING ON

PROMPT Gathering stats on user ${user} schema

EXEC DBMS_STATS.GATHER_TABLE_STATS('$user', 'CF1' );


EXIT;
EOF
ret=$?

if [ $ret -ne 0 ]
then
	msg FATAL
	msg FATAL "${FUNCNAME}: Failed to gather stats on $user schema"
	msg FATAL
fi

sql="ALTER TABLE cf1 NOPARALLEL"
if ( ! run_ddl $user $pass "$constring" "$sql" )
then
       msg FATAL "${FUNCNAME}: SQL execution failed in $user schema: \"${sql}\" "
       return 1
fi

return $ret
}

function grant() {
#global WORK_DIR
local user=$1
local tablespace=$2
local constring="$3" 
local ret=0
local fname="${WORK_DIR}/drop_users.sql"

if ( ! create_file $fname  )
then
        msg FATAL ""
        msg FATAL "${FUNCNAME}: Cannot create file: $fname "
        msg FATAL ""
        return 1
fi

msg NOTIFY "Creating user: $user "

sqlplus "$constring" <<EOF
WHENEVER SQLERROR EXIT 2;
SET TERMOUT ON
SET ECHO ON
PROMPT User grants for $user 
WHENEVER SQLERROR EXIT 2;
GRANT CONNECT TO $user IDENTIFIED BY $user;
GRANT DBA TO $user;
ALTER USER $user DEFAULT TABLESPACE $tablespace ;
EXIT;
EOF

ret=$?

if [ $ret -ne 0 ]
then
	msg FATAL
	msg FATAL "${FUNCNAME}: Failed to create $user schema or grants failed"
	msg FATAL
else
	# Leave behind a cleanup script
	echo "DROP USER $user CASCADE;" >> $fname 2>&1
fi

return $ret
}

function gen_basetable_sql() {
#global WORK_DIR
local tablespace=$1
local s=$2
local e=$3
local fname="${WORK_DIR}/$4"

if ( ! create_file $fname )
then
	msg FATAL ""
	msg FATAL "${FUNCNAME}: Cannot create file: $fname "
	msg FATAL ""
	return 1
fi


echo "
set termout on
set timing on
set echo on
WHENEVER SQLERROR EXIT 2;

CREATE TABLE keys_${s}(key NUMBER) PCTFREE 0 TABLESPACE $tablespace ;

DECLARE
	key NUMBER;
	x   VARCHAR2(128) := 'XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX';
	v_seed VARCHAR2(32) := TO_CHAR(SYSTIMESTAMP,'YYYYDDMMHH24MISSFFFF');

	CURSOR c IS SELECT key FROM keys_$s ORDER BY DBMS_RANDOM.VALUE();

BEGIN

	DBMS_RANDOM.seed (val => v_seed);

	FOR key IN ${s} .. ${e} LOOP
		INSERT /*+ APPEND */ INTO keys_$s  VALUES (key);
		IF ( MOD( key, 1023 ) = 0 ) THEN
			COMMIT;
		END IF;
        
	END LOOP;
	COMMIT;

	OPEN  c;
	LOOP
	FETCH c INTO key;
	EXIT WHEN c%NOTFOUND;
		INSERT /*+ APPEND */ INTO cf1 VALUES (key,x,x,x,x,x,x,x,x,x,x,x,x,x,x,x,x,x,x,x );
		IF ( MOD( key, 31 ) = 0 ) THEN
		COMMIT;
		END IF;
	END LOOP;
	COMMIT;
	CLOSE c;
END;
/

DROP TABLE keys_$s ;

EXIT;
" > $fname

if [ ! -r $fname ]
then
	msg FATAL ""
	msg FATAL "${FUNCNAME}: $fname is not a readable file"
	msg FATAL ""
	return 1
fi

return 0

}

function concurrent_table_load() {
#global LOG
#global SCALE
#global WORK_DIR
local tablespace=$1
local threads=$2
local user=$3
local pass=$4
local constring="$5"
local x=0
local b=0
local e=0
local tmp=0
local setsize=0
local flagfile="$WORK_DIR/.abort"
local sqlfile=""

if (  create_file $flagfile )
then
	rm -f $flagfile
else
	msg FATAL ""
        msg FATAL "${FUNCNAME}: Cannot create file: $fname "
	msg FATAL ""
        return 1
fi

(( setsize = $SCALE / $threads ))

for (( x=0 ; $x < $threads ; x++ ))
do
	(( b = ( $x * $setsize ) +  1  ))
	(( e = ( $x + 1 ) * $setsize ))
	# Deal with the remainder:
	[[ $x -eq $(( threads - 1 )) ]] && e=$SCALE

	sqlfile="ins_${x}.sql"
	if ( ! gen_basetable_sql $tablespace $b $e $sqlfile  >> $LOG 2>&1 )
	then
		msg FATAL ""
		msg FATAL "${FUNCNAME}: Concurrent load table setup failure"
		msg FATAL ""
		return 1	
	fi
done

# Now load the table 

tmp=$SECONDS
for (( x=0 ; $x < $threads ; x++ ))
do
	sqlfile="ins_${x}.sql"
	[[ ! -r $sqlfile ]] && msg FATAL "$sqlfile is not readable" && return 1
	( sqlplus ${user}/${pass}${constring}  @$sqlfile ; [[ $? -ne 0 ]] && touch $flagfile ;  rm -f $sqlfile ) &
done
wait

if [ -f $flagfile ]
then
	msg FATAL ""
	msg FATAL "${FUNCNAME}: Concurrent base table load failure"
	msg FATAL ""
	rm -f $flagfile
	return 1
else
	msg NOTIFY "User ${user} schema loaded in $(( SECONDS - tmp )) seconds"
fi

rm -f $flagfile

return 0
}

function create_index() {
#global UNIQUE_INDEX 
local user=$1
local pass=$2
local tablespace="$3"
local degree="$4"
local constring="$5"
local ret=0
local sql=""

sql="ALTER TABLE cf1 PARALLEL $degree "
if ( ! run_ddl $user $pass "$constring" "$sql" )
then
       msg FATAL "${FUNCNAME}: SQL execution failed in $user schema: \"${sql}\" "
       return 1
fi

sqlplus ${user}/${pass}${constring} <<EOF
SET TERMOUT ON
WHENEVER SQLERROR EXIT 2;
SET ECHO ON

PROMPT Creating index on ${user}.cf1 
CREATE ${UNIQUE_INDEX} INDEX I_CF1 ON cf1(custid) NOPARALLEL PCTFREE 0 TABLESPACE $tablespace;

ALTER INDEX i_cf1 SHRINK SPACE COMPACT;

EXIT;
EOF
ret=$?

sql="ALTER TABLE cf1 NOPARALLEL"
if ( ! run_ddl $user $pass "$constring" "$sql" )
then
       msg FATAL "${FUNCNAME}: SQL execution failed in $user schema: \"${sql}\" "
       return 1
fi

return $ret
}

function run_ddl(){
local user=$1
local pass=$2
local constring="$3"
local sql="$4"
local ret=0

sqlplus ${user}/${pass}${constring} <<EOF
WHENEVER SQLERROR EXIT 2;
SET ECHO ON
PROMPT $sql
$sql ;
EXIT;
EOF

ret=$?

if [ $ret -ne 0 ]
then

	msg FATAL "${FUNCNAME}: Failed to execute \"$sql\". Connect string: \"$user/${pass}${constring}\""
	return $ret
fi

return $ret
}

function setup_base_table () {
#global LOG
local user=$1
local pass=$2
local tablespace=$3
local threads=$4
local constring="$5"
local ret=0
local tmp=0

sqlplus ${user}/${pass}${constring} <<EOF
WHENEVER SQLERROR EXIT 2;
SET ECHO ON

PROMPT Forcing block sparseness on ${user}.cf1 
INSERT INTO cf1 VALUES (0, 'X', 'X', 'X', 'X', 'X', 'X', 'X', 'X', 'X', 'X', 'X', 'X', 'X', 'X', 'X', 'X', 'X', 'X', 'X');
COMMIT;

ALTER TABLE cf1 MINIMIZE RECORDS_PER_BLOCK;
TRUNCATE TABLE cf1 ;
COMMIT;
EXIT;
EOF

ret=$?

if [ $ret -ne 0 ]
then
	msg FATAL ""
	msg FATAL "${FUNCNAME}: Failed to create base table"
	msg FATAL ""
	return 1
fi

msg NOTIFY "Preparing to ${user} schema"

tmp=$SECONDS
if ( ! concurrent_table_load $tablespace $threads $user $pass "$constring" >> $LOG 2>&1 )
then
	msg FATAL ""
	msg FATAL "${FUNCNAME}: User ${user} table created but concurrent load procedure has failed"
	msg FATAL ""
	return 1
fi

msg NOTIFY "User ${user} data loading time: $(( SECONDS - tmp )) seconds"

tmp=$SECONDS
if ( ! create_index $user $pass $tablespace 64 "$constring")
then
	msg FATAL "${FUNCNAME}: Create index procedure failed for ${user} schema"
	return 1
fi

msg NOTIFY "User ${user} index creation time: $(( SECONDS - tmp )) seconds"

tmp=$SECONDS
if ! ( gather_stats $user $pass 64 "$constring" )
then
	msg FATAL "${FUNCNAME}: Stats gather procedure failed for ${user} schema"
	return 1

fi

msg NOTIFY "User ${user} CBO statistics gather time: $(( SECONDS - tmp )) seconds"

return 0
}

function load_normal_table() {
local user=$1
local pass=$2
local constring="$3"
local ret=0

sqlplus ${user}/${pass}${constring}  <<EOF
SET TERMOUT ON
WHENEVER SQLERROR EXIT 2;
SET ECHO ON

INSERT INTO cf1 SELECT * FROM user1.cf1 WHERE ROWNUM = 1 ;
COMMIT;

ALTER TABLE cf1 MINIMIZE RECORDS_PER_BLOCK;
TRUNCATE TABLE cf1 ;
COMMIT;

PROMPT Loading ${user}.cf1
INSERT /*+ APPEND */ INTO cf1 SELECT * FROM user1.cf1 ORDER BY DBMS_RANDOM.VALUE();
COMMIT;

EXIT;
EOF
ret=$?

return $ret
}

function create_table() {
local user=$1
local pass=$2
local tablespace=$3
local constring="$4"
local ret=0

sqlplus ${user}/${pass}${constring} <<EOF
SET TERMOUT ON
WHENEVER SQLERROR EXIT 2;
SET ECHO ON

REM Never set custid NOT NULL

PROMPT Creating table ${user}.cf1
CREATE TABLE cf1
(
custid NUMBER(15), c2 VARCHAR2(128), c3 VARCHAR2(128) ,
c4 VARCHAR2(128) , c5 VARCHAR2(128) , c6 VARCHAR2(128) ,
c7 VARCHAR2(128) , c8 VARCHAR2(128) , c9 VARCHAR2(128) ,
c10 VARCHAR2(128) , c11 VARCHAR2(128) , c12 VARCHAR2(128) ,
c13 VARCHAR2(128) , c14 VARCHAR2(128) , c15 VARCHAR2(128) ,
c16 VARCHAR2(128) , c17 VARCHAR2(128) , c18 VARCHAR2(128) ,
c19 VARCHAR2(128) , c20 VARCHAR2(128) ) 
PARALLEL CACHE PCTFREE 99 TABLESPACE $tablespace 
STORAGE (BUFFER_POOL RECYCLE INITIAL 1M NEXT 1M MAXEXTENTS UNLIMITED);
EXIT;
EOF
ret=$?

return $ret
}

function cr_slob_procedure() {
local user=$1
local pass=$2
local constring="$3"
local ret=0

msg NOTIFY "${PROGNAME}: Connecting via \"sqlplus ${user}/${pass}${constring}\" "

sqlplus ${user}/${pass}${constring} <<EOF
WHENEVER SQLERROR EXIT 2;
SET ECHO ON

CREATE OR REPLACE PROCEDURE slobupdate (pv_random PLS_INTEGER, pv_work_unit PLS_INTEGER, pv_redo_stress VARCHAR2) 
AUTHID current_user
AS BEGIN

IF ( pv_redo_stress = 'HEAVY' ) THEN
	UPDATE cf1 SET
	c2  =  'AAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBB',
	c3  =  'AAAAAAAABBBBBBBBAxAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBB',
	c4  =  'AAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAA5ABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBB',
	c5  =  'AAA0AAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBB4BBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBB',
	c6  =  'AAAAAArABBBBBBBBAAAAAAAABBBBBBBBAAAAAAtABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBB3BAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBB',
	c7  =  'AAA5AAAABBBBBBBtAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBB3BAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBB',
	c8  =  'AAAAAAAABBB0BBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBB3BAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBB',
	c9  =  'AAA0AAAABBBBBBBBAAAAAAAABrBBBBBBAAAAAArABBBBBBBBAAAAAAAABBB4BBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBB',
	c10  = 'AAAAAArABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBB3BAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBB',
	c11  = 'AAAAAAAABBBBBBBtAAAAAAAABBBBBBBBAAAAA-AABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBB3BAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBB',
	c12  = 'AAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBB3BAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBB',
	c13  = 'AAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAAB0BBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBB',
	c14 =  'AAAAAAAABBBBBBBBAAAAA4AABBBBBBBBAAAAAAAABBB9BBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBB3BBAAAAAAAABBBBBBBB',
	c15 =  'AAAAAAAABBBBBBBBAAAAAAAABB0BBBBBAAAAAAAABBBfBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBB3BBAAAAAAAABBBBBBBB',
	c16 =  'AAAAAAAABBBBBBBBAAAAAAAABBBBBBBBrAAAAAAABBBBBBBBAA9AAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBB3BBAAAAAAAABBBBBBBB',
	c17 =  'AAAAAAAABBBBBBBBAAA3AAAABBBBBBBBAAAAAAAABBBBBBBBAA-AAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBB3BBAAAAAAAABBBBBBBB',
	c18 =  '3AAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBB3BBAAAAAAAABBBBBBBB',
	c19 =  'A5AAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBB3BBAAAAAAAABBBBBBBB',
	c20 =  'AAAAAAAABBBBBBBBAAAAAAAA0BBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBB0BBBAAAAAAAABBBBBBBBAAAAAAAABBBBB3BBAAAAAAAABBBBBBBB'
	WHERE  custid >  ( pv_random - pv_work_unit ) AND  ( custid < pv_random);
	COMMIT;
ELSE
	UPDATE cf1 SET 
	c2  = 'AAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBB',
	c20 = 'AAAAAAAABBBBBBBBAAAAAAAA0BBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBBBBBBAAAAAAAABBBB0BBBAAAAAAAABBBBBBBBAAAAAAAABBBBB3BBAAAAAAAABBBB5BBB'
	WHERE  ( custid >  ( pv_random - pv_work_unit )) AND  (custid < pv_random);
	COMMIT;
END IF;
END slobupdate;
/
EXIT;
EOF

ret=$?
return $ret

}

function setup() {
#global LOG
local user=$1
local pass=$2
local tablespace=$3
local threads=$4
local constring="$5"

msg NOTIFY "Creating SLOB table for schema ${user}"

if ( ! create_table $user $pass $tablespace "$constring" >> $LOG 2>&1 )
then
	msg FATAL ""
	msg FATAL "${FUNCNAME}: Failed to create table for ${user}"
	msg FATAL ""
	return 1
fi

if [ "$user" = "user1" ]
then


	if ( ! setup_base_table $user $pass $tablespace $threads "$constring" >> $LOG 2>&1 )
	then
		msg FATAL ""
		msg FATAL "${FUNCNAME}: Failed to load ${user} SLOB table"
		msg FATAL ""
		return 1
	fi
else
	if ( ! load_normal_table $user $pass "$constring" >> $LOG 2>&1 )
	then
		msg FATAL ""
		msg FATAL "${FUNCNAME}: Failed to load ${user} SLOB table"
		msg FATAL ""
		return 1			
	fi

	if ( ! create_index $user $pass $tablespace 2 "$constring"  >> $LOG 2>&1 )
	then
		msg FATAL ""
		msg FATAL "${FUNCNAME}: Failed to create index for user ${user}"
		msg FATAL ""
		return 1			
	fi

	if ( ! gather_stats $user $pass 2 "$constring" >> $LOG 2>&1 )
	then
		msg FATAL ""
		msg FATAL "${FUNCNAME}: Failed to CBO stats for user ${user}"
		msg FATAL ""
		return 1			
	fi
fi

return 0
}

function create_log(){
local f=$1

if ( ! create_file $f )
then
	msg FATAL ""
	msg FATAL "${FUNCNAME}: Cannot create $f log file"
	msg FATAL ""
	return 1
fi

return 0
}

function pre_run_cleanup() {
#global ABORT_FLAG_FILE
local f=""

for f in drop_users.sql slob_data_load_summary.txt $ABORT_FLAG_FILE
do
	rm -f $f
done
return 0
}



#---------- Main body

export WORK_DIR=`pwd`
export LOG=${WORK_DIR=}/cr_tab_and_load.out
export ABORT_FLAG_FILE=$WORK_DIR/.abort_slob_load
export MAX_SLOB_SCHEMAS=4096
export MIN_SCALE=10000
tmp=""
user_scale_value=""

if [ $# -ne 2 ] 
then
	msg FATAL ""
	msg FATAL "Incorrect command line options"
	msg FATAL "Usage : ${0}: <tablespace name> <number of SLOB schemas to create and load>" 
	msg FATAL ""
	exit 1
else

	if ( ! is_int "$2" )
	then
		msg FATAL "Usage : ${0}: <tablespace name> <number of SLOB schemas to create and load>"
		msg FATAL "Option 2 must be an integer"
		exit 1
	fi

	export TABLESPACE="$1"
	export SCHEMAS="$2"

	if [[ "$SCHEMAS" -le 0 || "$SCHEMAS" -gt $MAX_SLOB_SCHEMAS ]] 
	then
		msg FATAL ""
		msg FATAL "Number of SLOB schemas must be integer and tested maximum is $MAX_SLOB_SCHEMAS"
		msg FATAL "Usage : ${0}: <tablespace name> <number of SLOB schemas to create and load>"
		msg FATAL ""
		exit 1
	fi
fi

if [ ! -r ./slob.conf ]
then
	msg FATAL "There is no readable slob.conf file in `pwd`"
	exit 1		
fi

msg NOTIFY ""
msg NOTIFY "Begin SLOB setup."

if ( ! create_log $LOG )
then
	msg FATAL ""
	msg FATAL "Cannot create log file (\"${LOG}\")"
	msg FATAL ""
	exit 1
fi

if ( ! test_oracle_utilities )
then
	msg FATAL ""
	msg FATAL "Abort. See ${LOG}"
	msg FATAL ""
	exit 1
fi

pre_run_cleanup 
source ./slob.conf

BLOCK_SZ=${BLOCK_SZ:=8192}
UNIQUE_INDEX=${UNIQUE_INDEX:=''}
LOAD_PARALLEL_DEGREE=${LOAD_PARALLEL_DEGREE:=1}
SCALE=${SCALE:=10000}
ADMIN_SQLNET_SERVICE=${ADMIN_SQLNET_SERVICE:=''}
SQLNET_SERVICE_BASE=${SQLNET_SERVICE_BASE:=''}
SYSDBA_PASSWD=${SYSDBA_PASSWD:=''}
export NON_ADMIN_CONNECT_STRING=" " # NOTE: This must either be a whitespace or a functioning value
export ADMIN_CONNECT_STRING=""

if [ -n "$ADMIN_SQLNET_SERVICE" ]
then

	if [ -z "$SYSDBA_PASSWD" ]
	then
		msg FATAL ""
		msg FATAL "slob.conf->SYSDBA_PASSWD is null"
		msg FATAL ""
		msg FATAL "If slob.conf->ADMIN_SQLNET_SERVICE is set then slob.conf->SYSDBA_PASSWD is required"
		msg FATAL "The connect convention for $0 via SQL*Net is \"sys/\${SYSDBA_PASSWD}\${ADMIN_SQLNET_SERVICE} as sysdba\""
		msg FATAL "ABORT"
		exit 1
	fi

	export ADMIN_CONNECT_STRING="sys/${SYSDBA_PASSWD}@${ADMIN_SQLNET_SERVICE} as sysdba"
	export NON_ADMIN_CONNECT_STRING="@${ADMIN_SQLNET_SERVICE}"

	msg NOTIFY "slob.conf->ADMIN_SQLNET_SERVICE is set to \"$ADMIN_SQLNET_SERVICE\". Testing listener for service status"

	if ( ! test_listener "$ADMIN_SQLNET_SERVICE" >> $LOG 2>&1 )
	then
		msg FATAL "slob.conf->ADMIN_SQLNET_SERVICE is set but tnsping cannot validate the service ($ADMIN_SQLNET_SERVICE)"
		msg FATAL "Please see ${LOG} for errors"
		exit 1
	fi
	
else
	export ADMIN_CONNECT_STRING="/ as sysdba"
fi

# Use fix_scale to convert any legal value to number of blocks
user_scale_value="$SCALE"
SCALE=$(fix_scale "$BLOCK_SZ" "$user_scale_value") ; ret=$?

if [ $ret -ne 0 ]
then
        msg FATAL "The value assigned to slob.conf->SCALE ($user_scale_value [$SCALE]) is an illegal value. "
	msg FATAL "Illegal SCALE value. Mininum value is $MIN_SCALE blocks"
        msg FATAL "Abort"
        exit 1
fi


msg NOTIFY "Load parameters from slob.conf: "
echo " 
SCALE: $user_scale_value ($SCALE blocks)
LOAD_PARALLEL_DEGREE: $LOAD_PARALLEL_DEGREE
ADMIN_SQLNET_SERVICE: \"$ADMIN_SQLNET_SERVICE\"
SQLNET_SERVICE_BASE: \"$SQLNET_SERVICE_BASE\"

Connect strings to be used:
ADMIN_CONNECT_STRING: \"$ADMIN_CONNECT_STRING\"
NON_ADMIN_CONNECT_STRING: \"$NON_ADMIN_CONNECT_STRING\"
"
msg NOTIFY "Testing connectivity to the instance to validate slob.conf settings"
msg NOTIFY "Testing Admin connect using \"$ADMIN_CONNECT_STRING\""

if ( ! test_conn "$ADMIN_CONNECT_STRING" >> $LOG 2>&1  )
then
	msg FATAL "Cannot connect to the instance"
	msg FATAL "Check $LOG log file for more information"
	msg FATAL "Please verify the instance is running and the settings"
	msg FATAL "in slob.conf are correct for your connectivity model"
	msg FATAL ""
	msg FATAL "If not connecting via SQL*Net please check \$ORACLE_SID"
	msg FATAL ""

	exit 1
else
	msg NOTIFY "Successful test connection: \"sqlplus -L $ADMIN_CONNECT_STRING\""
fi

if ( ! check_mto "$ADMIN_CONNECT_STRING")
then
	msg FATAL ""
	msg FATAL "This version of SLOB does not support Oracle Multitenant Option"
	msg FATAL

	exit 1
fi


USER=""
PASS=""

cnt=1
groupcnt=0
x=0
num_batches=0

if ( ! test_tablespace "$ADMIN_CONNECT_STRING" $TABLESPACE >> $LOG 2>&1 )
then
	msg FATAL "Cannot create tables in user-specified tablespace (\"$TABLESPACE\")"
	msg FATAL "See ${LOG}"
	exit 1
fi

msg NOTIFY "Dropping prior SLOB schemas. This may take a while if there is a large number of old schemas."
if ( ! drop_users "$ADMIN_CONNECT_STRING" )
then
	msg FATAL "Processing the DROP USER CACADE statements to remove any prior SLOB schemas failed."
	msg FATAL
	msg FATAL "Please check ${LOG}."
	msg FATAL ""
	exit 1
else
	msg NOTIFY "Previous SLOB schemas have been removed"
fi

msg NOTIFY "Preparing to load $SCHEMAS schema(s) into tablespace: $TABLESPACE"

if [ $SCHEMAS -gt 1 ]
then
	# this is not single schema but how many batches?
	(( num_batches = ( $SCHEMAS - 1 ) / $LOAD_PARALLEL_DEGREE ))
		
fi

while [ $cnt -le $SCHEMAS ]
do
	USER=user$cnt
	PASS=user$cnt

	if [ $cnt -eq 1 ]
	then
		if ( ! grant $USER $TABLESPACE "$ADMIN_CONNECT_STRING" >> $LOG  2>&1 )
		then
			msg FATAL "Cannot create ${USER} schema. See ${LOG}"
			exit 1
		fi

		msg NOTIFY "Loading $USER schema"
		before_load_ts=$SECONDS

		if ( ! setup $USER $PASS $TABLESPACE $LOAD_PARALLEL_DEGREE "$NON_ADMIN_CONNECT_STRING"  >> $LOG 2>&1 )
		then
			msg FATAL "Load procedure failed for $USER schema. See ${LOG}"
			exit 1
		fi

		before_concurrent_load="$SECONDS"

		msg NOTIFY "Finished loading, indexing and gathering statistics on user1 schema in $(( before_concurrent_load - before_load_ts )) seconds"
		[[  $SCHEMAS -gt 1 ]] && msg NOTIFY "Commencing multiple, concurrent schema creation and loading"
	else

		if ( ! grant $USER $TABLESPACE "$ADMIN_CONNECT_STRING" >> $LOG  2>&1 )
		then
			msg FATAL "Cannot create ${USER} schema. See ${LOG}"
			exit 1
		fi

		( setup $USER $PASS $TABLESPACE $LOAD_PARALLEL_DEGREE "$NON_ADMIN_CONNECT_STRING" >> $LOG 2>&1 || flag_abort $ABORT_FLAG_FILE ) &

		if [ $x -eq $(( LOAD_PARALLEL_DEGREE - 1 ))  ] 
		then
			before_load_ts=$SECONDS
			(( groupcnt = $groupcnt + 1 ))
			msg NOTIFY "Waiting for background batch ${groupcnt}. Loading up to user${cnt}"  
			wait 

			msg NOTIFY "Finished background batch ${groupcnt}. Load / index create / stats gather in $(( SECONDS - before_load_ts )) seconds"  
			x=0
		else
			before_load_ts=$SECONDS
			(( x = $x + 1 ))
		fi

		if ( check_abort_flag $ABORT_FLAG_FILE )
		then
			msg FATAL "Aborting SLOB setup. See ${LOG}"
			exit 1
		fi
	fi
	(( cnt = $cnt + 1 ))
done

if [ $SCHEMAS -gt 1 ]
then

	[[ $(( (SCHEMAS - 1)  % $LOAD_PARALLEL_DEGREE )) -ne 0  ]] && msg NOTIFY "Waiting for background batch $(( groupcnt + 1 )). Loading up to user${SCHEMAS}"

	wait

	(( concurrent_load_tm = $SECONDS - $before_concurrent_load ))
	if [[ $num_batches -gt 0  &&  $(( (SCHEMAS - 1)  % $LOAD_PARALLEL_DEGREE )) -ne 0 ]]
	then
		msg NOTIFY "Finished background batch $(( groupcnt + 1 )). Load / index create / stats gather in $(( SECONDS - before_load_ts )) seconds"  
			
	fi

	msg NOTIFY "Completed concurrent data loading phase: ${concurrent_load_tm} seconds"
fi

# Create the user0 schema for spare and compatibility with older SLOB
if ( ! grant user0 $TABLESPACE "$ADMIN_CONNECT_STRING" >> $LOG  2>&1 )
then
	msg FATAL "Cannot create ${USER} schema. See ${LOG}"
	exit 1
fi

msg NOTIFY "Creating SLOB procedure"

if ( ! cr_slob_procedure user1 user1 "$NON_ADMIN_CONNECT_STRING" >> $LOG 2>&1 )
then
	msg FATAL "Failed to create SLOB PL/SQL procedure"
	exit 1
fi

wait

msg NOTIFY "SLOB procedure created"

if ( slob_tabs_report "$ADMIN_CONNECT_STRING" >> $LOG 2>&1 )
then
	msg NOTIFY "Row and block counts for SLOB table(s) reported in ./slob_data_load_summary.txt"
	msg NOTIFY "Please examine ./slob_data_load_summary.txt for any possbile errors" 
else
	msg FATAL "Failed to generate SLOB table row and block count report"
	msg FATAL "See ${LOG}"
	exit 1
fi

msg NOTIFY ""
msg NOTIFY "NOTE: No errors detected but if ./slob_data_load_summary.txt shows errors then"
msg NOTIFY "examine ${LOG}"
echo ""

msg NOTIFY "SLOB setup complete. Total setup time:  (${SECONDS} seconds)"

check_wait_kit

exit 0

